/*
 * FtpClient.java
 *
 * Created on 10 novembre 2006, 11:18
 *

 */

package com.ewon.ewonlib.ftpclient;

import javax.microedition.io.*;
import java.io.*;
import javax.microedition.io.file.FileConnection;
import java.util.Vector;

/**
 * This class implements a Java FTP client.<br> <br>
 * It allows you to:
 * <ul> 
 * <li>connect to a FTP server in active and passive mode.</li>
 * <li>naviguate through the directories on the server</li>
 * <li>retrieve files (get) and store files (put) on the server</li> 
 * <li>delete files on the server</li>
 * <li>to make and remove directories on the server</li>
 * </ul>
 *
 *<H3>Example:</H3>
 *<PRE>
 *     ftpClient = new FtpClient();
 *     try {
 *        ftpClient.connect("10.0.100.206",21,"adm","adm");
 *        ftpClient.setActivePortRange(10000,20000);
 *        ftpClient.setActive();
 *        String[] dirList = ftpClient.dir("");
 *        for (int i=0; i< dirList.length; i++) {
 *           System.out.println(dirList[i]);
 *        }
 *     } catch (IO Exception io) {
 *        System.out.println ("Handle IOException caught:"+e);
 *     } catch (FtpCException f) {
 *        System.out.println ("Handle FtCException caught:"+f);
 *     } finally {
 *        ftpClient.disconnect();
 *     }
 *</PRE>
 * @author ahe
 */


public class FtpClient {
    
    
    /**
     * Create an instance of FtpClient.
     */
    public FtpClient() {
    }
    
    /**
     * Connects to the default port of an FTP server and logs in as anonymous/anonymous.
     * @param        host hostname or IP address of the server
     * @exception    <code>IOException</code>  thrown by external used classes
     * @exception    <code>FtpCException</code> thrown whea non valid reply has been returned by the FTP server or when already connected.   
     */
    public synchronized void connect(String host) throws  IOException,FtpCException {
        connect(host, 21);
    }
    
    /**
     * Connects to an FTP server and logs in as anonymous/anonymous.
     * @param        host server hostname or server IP address
     * @param        Port server port 
     * @exception    <code>IOException</code>  thrown by external used classes
     * @exception    <code>FtpCException</code> thrown whea non valid reply has been returned by the FTP server or when already connected.   
     */

    public synchronized void connect(String host, int Port) throws IOException ,FtpCException {
        connect(host, Port, "anonymous", "anonymous");
    }
    
    /**
     * Connects to an FTP server and logs in with the supplied username and password.
     * @param        host server hostname or server IP address
     * @param        Port server port
     * @param        user username
     * @param        pass user password 
     * @exception    <code>IOException</code>  thrown by external used classes
     * @exception    <code>FtpCException</code> thrown whea non valid reply has been returned by the FTP server or when already connected.   
     */

    public synchronized void connect(String host, int Port, String user, String pass) throws IOException,FtpCException {
        
       String openParm = "socket://" + host + ":" + Port;

       if (control != null) {
           throw new FtpCException("FtpClient is already connected. Disconnect first");
       }
       
       hostAddr = new String(host);
       
       try {
          control  = new FtpControlSocket();
          control.open(openParm);
 
          String[] validCodes = {"220"};
          readAndValidateReply(validCodes);
     
          String command = "USER " + user+ "\r\n";
          control.write(command.getBytes());
          String[] validReplyCodes = {"230","331"};
          readAndValidateReply(validReplyCodes);

          command = "PASS " + pass + "\r\n";
          control.write(command.getBytes());
          String[] validReplyCodes2 = {"230","202"};
          readAndValidateReply(validReplyCodes2);
       } catch (IOException io) {
          try {
             control.close();
          } catch (IOException e) {
          } finally {
             control = null;
             throw io;
          } 
       } 
        // Now logged in.
    }
    
    /**
     * Disconnects from the FTP server. 
     */
    public synchronized void disconnect() {
    
       if (control == null) {
           return;
       }
        String command  = "QUIT\r\n";
        try {
            control.write(command.getBytes());
        } catch (IOException e){
        }
        catch (FtpCException f) {
        }
        finally {  
            if (control != null) {
               try {
               control.close();
               } catch (IOException e) {
               } finally {
               control = null;
               }
            }
        }
    }
    
     /**
     * Sets passive mode: data connections will be set up in passive mode.
     * @exception    <code>FtpCException</code> thrown when not connected to the FTP server.
     */
    public synchronized void setPassive() throws FtpCException {
        
       if (control == null) {
           throw new  FtpCException("FtpClient is not connected. Connect first");
       } 
       activeMode = false;

    }
    
     /**
     * Sets the port range to use when data connections are set up in active mode.
     * @param        Lowest indicate lowest port of the port range
     * @param        Highest indicate highest port of the port range
     * @exception    <code>FtpCException</code> thrown when an invalid port range is specified. 
     */
    
    public synchronized void setActivePortRange(int Lowest, int Highest)  throws FtpCException {
        if (Lowest < 0 || Lowest > Highest || Highest > 65536) {
            throw new FtpCException("Invalid port range specified");
        }   
        activeLowPort  = Lowest;
        activeHighPort = Highest; 

    }
    
     /**
     * Sets active mode: data connections will be set up 
     * in active mode.
     * @exception  <code>FtpCException</code> thrown when not connected to the FTP server.
     */   
    public synchronized boolean setActive() throws FtpCException {
       if (control == null) {
           throw new FtpCException("FtpClient is not connected. Connect first");
       }
       activeMode = true;
       return true;
    }
    
     /**
     * Changes working directory.
     * @param      dir destination directory
     * @exception    <code>IOException</code>  thrown by external used classes
     * @exception    <code>FtpCException</code> thrown whea non valid reply has been returned by the FTP server or when not connected.   
     */

    public synchronized void cwd(String dir) throws IOException,FtpCException {
       if (control == null) {
           throw new FtpCException("FtpClient is not connected. Connect first");
       }  
       try  {
          String command = "CWD " + dir + "\r\n";
          control.write(command.getBytes());
          String[] validCodes = {"250","230"};
          String response = readAndValidateReply(validCodes);
          if (response.startsWith("230")) {
              String[] validReplyCodes2 = {"250"};
              response = readAndValidateReply(validReplyCodes2);
          }
       } catch (IOException io) {
          try {
             control.close();
          } catch (IOException e) {
          } finally {
             control = null;
             throw io;
          } 
       } 
    }
    
     /**
     * Returns the present working directory.
     * @return     current working directory
     * @exception    <code>IOException</code>  thrown by external used classes
     * @exception    <code>FtpCException</code> thrown whea non valid reply has been returned by the FTP server or when not connected.   
     */
    public synchronized String pwd() throws  IOException,FtpCException {

       if (control == null) {
           throw new FtpCException("FtpClient is not connected. Connect first");
       }  
       String response = null;
       try {
          String command = "PWD\r\n";
          control.write(command.getBytes());
          String[] validCodes = {"257","230"};
          response = readAndValidateReply(validCodes);
          if (response.startsWith("230")) {
              String[] validReplyCodes2 = {"257"};
              response = readAndValidateReply(validReplyCodes2);
          }
       } catch (IOException io) {
          try {
             control.close();
          } catch (IOException e) {
          } finally {
             control = null;
             throw io;
          } 
       } 
       String dir     = null;
       int firstQuote  = response.indexOf('\"');
       int secondQuote = response.indexOf('\"',firstQuote +1);
       if (secondQuote >0) {
           dir = response.substring(firstQuote+1,secondQuote);
       }        
       return dir;
    }
    
     /**
     * Creates a new directory.
     * @param      dir directory name of the directory to be created
     * @exception    <code>IOException</code>  thrown by external used classes
     * @exception    <code>FtpCException</code> thrown whea non valid reply has been returned by the FTP server or when not connected.   
     */
    public synchronized void mkd(String dir) throws IOException,FtpCException {
     
       if (control == null) {
           throw new FtpCException("FtpClient is not connected. Connect first");
       }
       try {
          String command = "MKD " + dir + "\r\n";
          control.write(command.getBytes());
          String[] validCodes = {"257","200","230"};
          String response = readAndValidateReply(validCodes);
          if (response.startsWith("230")) {
              String[] validReplyCodes2 = {"257","200"};
              response = readAndValidateReply(validReplyCodes2);
          }
       } catch (IOException io) {
         try {
            control.close();
          } catch (IOException e) {
          } finally {
             control = null;
              throw io;
          } 
       }     
    }
     /**
     * Deletes a directory
     * @param      dir directory name of the directory to be removed
     * @exception    <code>IOException</code>  thrown by external used classes
     * @exception    <code>FtpCException</code> thrown whea non valid reply has been returned by the FTP server or when not connected.   
     */
    public synchronized void rmd(String dir) throws IOException,FtpCException {
      if (control == null) {
          throw new FtpCException("FtpClient is not connected. Connect first");
      } 
      try {
         String command = "RMD " + dir + "\r\n";
         control.write(command.getBytes());
         String[] validCodes = {"250","200","230"};
         String response = readAndValidateReply(validCodes);
         if (response.startsWith("230")) {
              String[] validReplyCodes2 = {"250","200"};
              response = readAndValidateReply(validReplyCodes2);
          }
      } catch (IOException io) {
         try {
            control.close();
         } catch (IOException e) {
         } finally {
            control = null;
            throw io;
         } 
       }     
    }
     /**
     * Triggers ok reply from server side
     * @exception    <code>IOException</code>  thrown by external used classes
     * @exception    <code>FtpCException</code> thrown whea non valid reply has been returned by the FTP server or when not connected.   
     */
    public synchronized  void noop() throws  IOException,FtpCException {

       if (control == null) {
           throw new FtpCException("FtpClient is not connected. Connect first");
       }  
       String response = null;
       try {
          String command = "NOOP\r\n";
          control.write(command.getBytes());
          String[] validCodes = {"200","230"};
          response = readAndValidateReply(validCodes);
          if (response.startsWith("230")) {
              String[] validReplyCodes2 = {"200"};
              response = readAndValidateReply(validReplyCodes2);
          }
       } catch (IOException io) {
          try {
             control.close();
          } catch (IOException e) {
          } finally {
             control = null;
             throw io;
          } 
       }        
       return;
    }
    
     /**
     * Stores a file on the remote server
     * @param      local file name of the file to be stored on the server
     * @param      remote file name on the server of the file to be stored
     * @exception    <code>IOException</code>  thrown by external used classes
     * @exception    <code>FtpCException</code> thrown whea non valid reply has been returned by the FTP server or when not connected.   
     */
    public synchronized void put(String local, String remote) throws IOException,FtpCException {
       FileConnection fconn = null;
       InputStream fis = null;
       String urlFileName = "file:///"+local;
       fconn = (FileConnection) Connector.open(urlFileName);
       if (!fconn.exists()) {
               throw new FtpCException("file to transfer does  not exist.");
       }
       fis = fconn.openInputStream();
       try {
          put (fis,remote);
       } finally {
          try {
             fis.close();
          } finally {
             fconn.close();
          }
       }
    }
    
     /**
     * Stores an input stream on the remote server
     * @param      iStream input stream to be stored
     * @param      remote file name on the server of the file to be stored
     * @exception  <code>IOException</code>  thrown by external used classes
     * @exception  <code>FtpCException</code> thrown whea non valid reply has been returned by the FTP server or when not connected.   
     */
    public synchronized void put(InputStream iStream, String remote) throws IOException,FtpCException {
        
       FtpDataSocket             dataSocket = null; 
       if (control == null) {
          throw new FtpCException("FtpClient is not connected. Connect first");
       } 
       if (activeMode == false) {
           setPassiveMode(); 
       } else {
           setActiveMode();
       }
       dataSocket = new FtpDataSocket();
       if (activeMode == false) {
           String OpenDataParm = "socket://" + hostAddr + ":" + passiveDataPort; 
           dataSocket.openPassive(OpenDataParm);
       } else {
           String OpenDataParm = "socket://" + ":" + activeDataPort; 
           dataSocket.openActive(OpenDataParm);
           /* if (activeMode == true) {
             dataSocket.acceptActive();
            
             }   */ 
       }
       try {
          String command     = "STOR " + remote + "\r\n";
          control.write(command.getBytes());
       } catch (IOException io) {
         try {
            control.close();
         } catch (IOException e) {
         } finally {
            control = null;
            throw io;
         } 
       }           
       

       try {
          String[] validCodes = {"150","125"};
          readAndValidateReply(validCodes);
       } catch (IOException io) {
         try {
            try {
               dataSocket.close();
            } finally {
               control.close();
            }
         } catch (IOException e) {
         } finally {
            control = null;
            throw io;
         } 
       }  
       if (activeMode == true) {
          dataSocket.acceptActive();
       }
       try {
          int bytesRead;
          int length;
          int avail;
          length = 1024;
          byte[] buffer = new byte[length];
           // ne retourne pas -1 si fin de fichier (end of stream))
          while ((avail = iStream.available())!=0) {  
             if (avail > buffer.length) {
                 avail =buffer.length;
             }
             bytesRead = iStream.read(buffer,0,avail);
             dataSocket.write(buffer,0,bytesRead);
          }
       } finally {
          dataSocket.close();
       }
       try {
          String[] validReplyCodes = {"226"};
          readAndValidateReply(validReplyCodes);
       }  catch (IOException io) {
         try {
            control.close();
         } catch (IOException e) {
         } finally {
            control = null;
            throw io;
         } 
       }
    }
    /**
     * gets a file from the remote server
     * @param      local local file name of the file to be retrieved
     * @param      remote file name on the server of the file to be retrieved
     * @exception    <code>IOException</code>  thrown by external used classes
     * @exception    <code>FtpCException</code> thrown whea non valid reply has been returned by the FTP server or when not connected.   
     */
    public synchronized void get(String local, String remote) throws IOException,FtpCException {
       FileConnection fconn = null;
       OutputStream   fos   = null;
       String urlFileName = "file:///"+ local;
       fconn = (FileConnection) Connector.open(urlFileName);
       if (!fconn.exists())
          fconn.create();  // create the file if it doesn't exist
       fos = fconn.openOutputStream();
       try {
           get(fos,remote);
       } finally {
         try {
            fos.close();
         } finally {
            fconn.close();
         }
       }
    }
    
     /**
     * gets a file from the remote server
     * @param      oStream output stream for the retrieved file
     * @param      remote file name on the server of the file to be retrieved
     * @exception    <code>IOException</code>  thrown by external used classes
     * @exception    <code>FtpCException</code> thrown whea non valid reply has been returned by the FTP server or when not connected.   
     */
    
    public synchronized void get(OutputStream oStream, String remote) throws IOException,FtpCException {
        
       FtpDataSocket             dataSocket = null;
       if (control   == null) {
           throw new FtpCException("FtpClient is not connected. Connect first");
       }
       if (activeMode == false) {
           setPassiveMode(); 
       } else {
           setActiveMode();
       }
       dataSocket = new FtpDataSocket();
       if (activeMode == false) {
          String OpenDataParm = "socket://" + hostAddr + ":" + passiveDataPort; 
          dataSocket.openPassive(OpenDataParm);
       }
       else {
          String OpenDataParm = "socket://" + ":" + activeDataPort; 
          dataSocket.openActive(OpenDataParm);     
       }
       try {
          String command = "RETR " + remote + "\r\n";
          control.write(command.getBytes());
       } catch (IOException io) {
         try {
            control.close();
         } catch (IOException e) {
         } finally {
            control = null;
            throw io;
         } 
       }

       
       try {
          String[] validCodes = {"150","125"};
          readAndValidateReply(validCodes);
       } catch (IOException io) {
         try {
            try {
               dataSocket.close();
            } finally {
               control.close();
            }
         } catch (IOException e) {
         } finally {
            control = null;
            throw io;
         } 
       }
       if (activeMode == true) {
          dataSocket.acceptActive();
       }
       try {
          int bytesRead;
          int length;
          length = 1024;
          byte[] buffer = new byte[length];
          while ((bytesRead = dataSocket.read(buffer,0,length)) != -1) {
             oStream.write(buffer,0,bytesRead);
          }
       }  finally {
          dataSocket.close();
       }
       try { 
          String[] validReplyCodes = {"226"};
          readAndValidateReply(validReplyCodes);
       } catch (IOException io) {
         try {
            control.close();
         } catch (IOException e) {
         } finally {
            control = null;
            throw io;
         } 
       }
    } 
     /**
     * Deletes a file 
     * @param      fileName file name of the file to be deleted
     * @exception    <code>IOException</code>  thrown by external used classes
     * @exception    <code>FtpCException</code> thrown whea non valid reply has been returned by the FTP server or when not connected.   
     */
    
    public synchronized void delete(String fileName) throws IOException,FtpCException {
        
       if (control == null) {
           throw new FtpCException("FtpClient is not connected. Connect first");
       }
       try {
          String command = "DELE " + fileName + "\r\n";
          control.write(command.getBytes());
          String[] validCodes = {"250","200","230"};
          String response = readAndValidateReply(validCodes);
          if (response.startsWith("230")) {
              String[] validReplyCodes2 = {"250","200"};
              response = readAndValidateReply(validReplyCodes2);
          }
       } catch (IOException io) {
         try {
            control.close();
         } catch (IOException e) {
         } finally {
            control = null;
            throw io;
         } 
       }
    }
     /**
     * Lists the contents of a directory
     * @param      dir name of the directory
     * @return     files and directories (name only) of the specified directory
     * @exception    <code>IOException</code>  thrown by external used classes
     * @exception    <code>FtpCException</code> thrown whea non valid reply has been returned by the FTP server or when not connected.   
     */

    public String[] dir(String dir) throws IOException,FtpCException {
        
        return dir(dir,"",false);
    }
     /**
     * Lists the contents of a directory according to specified attributes
     * @param      dir name of the directory
     * @param      attr ls command like attributes
     * @return     files and directories of the specified directory
     * @exception    <code>IOException</code>  thrown by external used classes
     * @exception    <code>FtpCException</code> thrown whea non valid reply has been returned by the FTP server or when not connected.   
     */
   

    public String[] dir(String dir, String attr) throws IOException,FtpCException {
        
        return dir(dir,attr,true);
    }
     /**
     * Lists the contents of a directory according to specified attributes
     * @param      dir name of the directory
     * @param      attr ls command like attributes
     * @return     files and directories of the specified directory
     * @exception    <code>IOException</code>  thrown by external used classes
     * @exception    <code>FtpCException</code> thrown whea non valid reply has been returned by the FTP server or when not connected.   
     */

    private String[] dir(String dir, String attr, boolean useList) throws IOException,FtpCException {
        
       FtpDataSocket   dataSocket = null; 

       if (control == null) {
           throw new FtpCException("FtpClient is not connected. Connect first");
       }
       if (activeMode == false) {
           setPassiveMode(); 
       } else {
           setActiveMode();
       }
       
       dataSocket = new FtpDataSocket();
       if (activeMode == false) {
           String OpenDataParm = "socket://" + hostAddr + ":" + passiveDataPort; 
           dataSocket.openPassive(OpenDataParm);
       } else {
           String OpenDataParm = "socket://" + ":" + activeDataPort; 
           dataSocket.openActive(OpenDataParm);
 
       }

       try {
          String command;
          if (!useList) {
             command = "NLST " + dir + "\r\n";
          }
          else {
             command = "LIST " + ((attr.length() != 0) ? (attr + " ") : "") + dir + "\r\n";   
          }
          control.write(command.getBytes()); 
       } catch (IOException io) {
         try {
            try {
               dataSocket.close();
            } finally {
               control.close();
            }
         } catch (IOException e) {
         } finally {
            control = null;
            throw io;
         } 
       }

       try {
          String[] validCodes = {"150","125"};
          readAndValidateReply(validCodes);
       }  catch (IOException io) {
         try {
            try {
               dataSocket.close();
            } finally {
               control.close();
            }
         } catch (IOException e) {
         } finally {
            control = null;
            throw io;
         } 
       }
       
       if (activeMode == true) {
          dataSocket.acceptActive();
       }
       
       Vector lines = new Vector();
       StringBuffer  dataStr = new StringBuffer();
       String line = null;
       String[] list    = new String [0];
       try {
          int  prevch,ch;
          ch='\n';
          prevch = 0;
          while (ch=='\n') {
             while (((ch = dataSocket.read()) != -1) && (!((prevch =='\r')&& (ch =='\n')))) {
                dataStr.append((char) ch); 
                prevch = ch;
             }
             if (ch!=-1) {
                 dataStr.setLength(dataStr.length()-1);
                 lines.addElement(dataStr.toString()); 
                 dataStr.setLength(0);
             }
          }
          if (!lines.isEmpty()) {
             list = new String[lines.size()];
             lines.copyInto(list);
          }
       } finally {
          dataSocket.close();
       }
       try {
          String[] validReplyCodes = {"226"};
          readAndValidateReply(validReplyCodes);
       } catch (IOException io) {
         try {
            control.close();
         } catch (IOException e) {
         } finally {
            control = null;
            throw io;
         } 
       }
       return list;
    }
    
    /**
    * Sets Binary mode 
     * @exception    <code>IOException</code>  thrown by external used classes
     * @exception    <code>FtpCException</code> thrown whea non valid reply has been returned by the FTP server or when not connected.   
     */
    
    public void setBin() throws IOException,FtpCException {  
       if (control== null) {
           throw new FtpCException("FtpClient is not connected. Connect first");
       }
       try {
          String command = "TYPE I\r\n";
          control.write(command.getBytes());
          String[] validReplyCodes = {"200","230"};
          String response = readAndValidateReply(validReplyCodes);
          if (response.startsWith("230")) {
              String[] validReplyCodes2 = {"200"};
              response = readAndValidateReply(validReplyCodes2);
          }
       } catch (IOException ex) {
         try {
            control.close();
         } catch (IOException e) {
         } finally {
            control = null;
            throw ex;
         } 
       }
    }
     /**
     * get current version of java FTP client library
     * @return     current version of java FTP client library
     */
    
    public String getVersion()
    {
       return libVersion;
    }
    
    private void setActiveMode() throws IOException,FtpCException {
        
       if (control== null) {
           throw new FtpCException("FtpClient is not connected. Connect first");
       } 
       StringBuffer localIP = new StringBuffer();
       
       try {
          String       localIPStr = control.getLocalAddress();
          localIP.append(control.getLocalAddress());
       
          int          ipIdx,ipIdxPrev;
          ipIdxPrev =0;
          for (int i=0;i<3;i++) {
             ipIdx = localIPStr.indexOf ('.',ipIdxPrev +1);
             localIP.setCharAt(ipIdx,',');
             ipIdxPrev = ipIdx;
          }
          int port = getActivePort();
          int portLowByte  = port % 256;
          int portHighByte = port / 256;
          String command = "PORT "+ localIP + "," + String.valueOf(portHighByte) +
                        "," + String.valueOf(portLowByte)+"\r\n";
          control.write(command.getBytes());
          String[] validReplyCodes = {"200","230"};
          String response = readAndValidateReply(validReplyCodes);
          if (response.startsWith("230")) {
              String[] validReplyCodes2 = {"200"};
              readAndValidateReply(validReplyCodes2);
          }
       } catch (IOException ex) {
         try {
            control.close();
         } catch (IOException e) {
         } finally {
            control = null;
            throw ex;
         } 
       }

    }
    private void setPassiveMode() throws IOException,FtpCException {
           
       if (control == null) {
           throw new FtpCException("FtpClient is not connected. Connect first");
       }  
       try {
          String command = "PASV\r\n";
          control.write(command.getBytes());
          String[] validReplyCodes = {"227","230"};
          String response = readAndValidateReply(validReplyCodes);
          if (response.startsWith("230")) {
              String[] validReplyCodes2 = {"227"};
              response = readAndValidateReply(validReplyCodes2);
          }
          int opening = response.indexOf ('(');
          int closing = response.indexOf (')',opening +1);
          int ipIdx = -1;
          int ipIdxPrev;
            /* ?? we don't have Tokenizer class  in J2ME ?? */
          if (closing > 0) {
             ipIdxPrev = opening;
             ipIdx     = -1;
             for (int i=0;i<4;i++) {
                ipIdx = response.indexOf (',',ipIdxPrev +1);
                if (ipIdx > 0)  {
                   ipIdxPrev = ipIdx;
                   continue;
                } else {
                   break;
                }
             } 
             if (ipIdx == ipIdxPrev) {
                ipIdx = response.indexOf(',',ipIdx+1);
                if (ipIdx >0) {

                   passiveDataPort = Integer.parseInt(response.substring(ipIdxPrev +1,ipIdx)) *256
                                  + Integer.parseInt(response.substring(ipIdx+1,closing));
                }
             }
          } 
       } catch (IOException ex) {
         try {
            control.close();
         } catch (IOException e) {
         } finally {
            control = null;
            throw ex;
         } 
       }
    }

    
    private String readAndValidateReply(String[] validReplies) throws IOException,FtpCException {
        
        StringBuffer str = new StringBuffer();
        int          prevch,ch;
        
        str.setLength(0);
        prevch = 0;
        
        while (((ch = control.read()) != -1) && (!((prevch =='\r')&& (ch =='\n')))) {
           str.append((char) ch); 
           prevch = ch;
        }
        String response = str.toString();
        for (int i = 0; i < validReplies.length; i++) {
           if (response.startsWith(validReplies[i])) {
                 return response;
           }
        }    
             // got this far, not recognised
        throw new FtpCException("Invalid ftp reply received: "+ response); 
        // return false;
    }
    
    private int getActivePort() {  
        if (activeDataPort <activeLowPort) {
            activeDataPort = activeLowPort; 
        } else {
           activeDataPort = activeDataPort +1;
           if (activeDataPort > activeHighPort) {
              activeDataPort = activeLowPort;
           }
        }
        return activeDataPort;
    }
    // ??? est ce que la mthode finalize  a un sens  ou deprecated ??? */

    private FtpControlSocket control;
    private boolean          activeMode = true;
    private int              activeLowPort   = 0;
    private int              activeHighPort  = 65536;
    private int              activeDataPort  = 0;
    private int              passiveDataPort = -1;
    private String           hostAddr = null;
    private String           libVersion ="1.1";
}